import enum

@enum.unique
class ErdOperatingMode(enum.Enum):
    LOW_POWER = 0
    POWER_UP = 1
    STANDBY = 2
    DELAY_START = 3
    PAUSE = 4
    CYCLE_ACTIVE = 5
    EOC = 6
    DOWNLOAD_MODE = 7
    SENSOR_CHECK_MODE = 8
    LOAD_ACTIVATION_MODE = 9
    MC_ONLY_MODE = 17
    WARNING_MODE = 18
    CONTROL_LOCKED = 19
    CSM_TRIPPED = 20
    INVALID = 255




import asyncio
import logging
from typing import Any, Dict, Optional, Union

import slixmpp
from lxml import etree
from ..erd import ErdCode, ErdCodeType
from ..exception import *
from ..ge_appliance import GeAppliance
from .base_client import GeBaseClient
from .const import *
from .ge_client_xmpp import (
    GeClientXMPP, 
    XMPP_EVENT_MESSAGE, 
    XMPP_EVENT_PRESENCE_AVAILABLE, 
    XMPP_EVENT_PRESENCE_UNAVAILABLE, 
    XMPP_EVENT_SESSION_START
)
from .states import GeClientState

try:
    import ujson as json
except ImportError:
    import json

_LOGGER = logging.getLogger(__name__)

def _first_or_none(lst: list) -> Any:
    try:
        return lst[0]
    except IndexError:
        return None

class GeXmppClient(GeBaseClient):
    client_priority = 1

    def __init__(self, username: str, password: str, event_loop: Optional[asyncio.AbstractEventLoop]):
        super().__init__(username, password, event_loop=event_loop)
        self._client = None  # type: Optional[GeClientXMPP]

    def _get_xmpp_client(self) -> GeClientXMPP:
        client = GeClientXMPP(self.credentials['jid'], self.credentials['password'])
        client.add_external_event_handler(XMPP_EVENT_SESSION_START, self._on_start)
        client.add_external_event_handler(XMPP_EVENT_MESSAGE, self._on_message)
        client.add_external_event_handler(XMPP_EVENT_PRESENCE_AVAILABLE, self._on_presence_available)
        client.add_external_event_handler(XMPP_EVENT_PRESENCE_UNAVAILABLE, self._on_presence_unavailable)

        return client

    async def _async_do_full_login_flow(self) -> Dict[str,str]:
        """Perform a complete login flow, returning credentials."""

        _LOGGER.debug('Getting OAuth2 token')
        await self._async_get_oauth2_token()

        _LOGGER.debug('Getting mobile device token')
        mdt = await self._async_get_mobile_device_token()

        _LOGGER.debug('Getting GE token')
        ge_token = await self._async_get_ge_token(mdt)

        _LOGGER.debug('Getting XMPP credentials')
        return await self._async_get_xmpp_credentials(ge_token)
    
    async def _async_do_refresh_login_flow(self) -> Dict[str, str]:
        """Perform a refresh login flow, returning credentials"""

        _LOGGER.debug('Refreshing OAuth2 token')
        await self._async_refresh_oauth2_token()

        _LOGGER.debug('Getting mobile device token')
        mdt = await self._async_get_mobile_device_token()

        _LOGGER.debug('Getting GE token')
        ge_token = await self._async_get_ge_token(mdt)

        _LOGGER.debug('Getting XMPP credentials')
        return await self._async_get_xmpp_credentials(ge_token)

    async def _async_get_mobile_device_token(self) -> str:
        """Get a mobile device token"""
        await self._set_state(GeClientState.AUTHORIZING_CLIENT)

        mdt_data = {
            'kind': 'mdt#login',
            'app': OAUTH2_APP_ID,
            'os': 'google_android'
        }
        auth_header = { 'Authorization': 'Bearer ' + self._access_token }

        async with self._session.post(f'{API_URL}/v1/mdt', json=mdt_data, headers=auth_header) as resp:
            if resp.status != 200:
                raise GeAuthFailedError(await resp.text())
            results = await resp.json()
        try:
            return results['mdt']
        except KeyError:
            raise GeAuthFailedError(f'Failed to get a mobile device token: {results}')

    async def _async_get_ge_token(self, mobile_device_token: str) -> str:
        """Get the GE token that we can use to get XMPP credentials"""
        params = {
            'client_id': OAUTH2_CLIENT_ID,
            'client_secret': OAUTH2_CLIENT_SECRET,
            'mdt': mobile_device_token
        }
        auth_header = { 'Authorization': 'Bearer ' + self._access_token }

        async with self._session.post(f'{LOGIN_URL}/oauth2/getoken', params=params, headers=auth_header) as resp:
            if 400 <= resp.status < 500:
                raise GeAuthFailedError(await resp.text())
            if resp.status >= 500:
                raise GeGeneralServerError(await resp.text())
            results = await resp.json()

        try:
            return results['access_token']
        except KeyError:
            raise GeAuthFailedError(f'Failed to get a GE token: {results}')

    async def _async_get_xmpp_credentials(self, ge_token: str) -> Dict:
        """Get XMPP credentials"""
        uri = f'{API_URL}/v1/mdt/credentials'
        headers = {'Authorization': f'Bearer {ge_token}'}
        async with self._session.get(uri, headers=headers) as resp:
            if 400 <= resp.status < 500:
                raise GeAuthFailedError(await resp.text())
            if resp.status >= 500:
                raise GeGeneralServerError(await resp.text())
            return await resp.json()

    async def _async_run_client(self):
        """Run the client."""
        try:
            await self._set_state(GeClientState.CONNECTING)
            address = (self._credentials['address'], self._credentials['port'])
            self._client = self._get_xmpp_client()
            self._client.connect(address=address)
            #run the loop
            await asyncio.ensure_future(self._client.disconnected, loop=self.loop)
        #TODO: better exception handling
        except Exception as err:
            _LOGGER.error(f"Exception while processing XMPP loop: {err}")
        finally:
            self._disconnect()

    async def _disconnect(self) -> None:
        if self._client:
            await self._client.disconnect()
        self._client = None

    async def _add_appliance(self, jid: str):
        """Add an appliance to the registry and request an update."""
        mac_addr = jid.split('_')[0]
        if jid in self.appliances:
            raise GeDuplicateApplianceError(f'Attempted to add duplicate appliances {mac_addr} ({jid})')
        new_appliance = GeAppliance(mac_addr, self)

        _LOGGER.info(f'Adding appliance {jid}')
        self.appliances[jid] = new_appliance
        await self.async_event(EVENT_ADD_APPLIANCE, new_appliance)
        await new_appliance.async_request_update()

    async def _maybe_add_appliance(self, jid: str):
        """Add an appliance, suppressing the error if it already exists."""
        try:
            await self._add_appliance(jid)
        except GeDuplicateApplianceError as err:
            _LOGGER.warn(f'{err}')
            pass

    async def _on_presence_available(self, evt: slixmpp.ElementBase):
        """Perform actions when notified of an available JID."""
        await asyncio.sleep(2)  # Wait 2 seconds to give it time to register
        jid = slixmpp.JID(evt['from']).bare

        if jid == self._client.boundjid.bare:
            return
        try:
            await self._set_appliance_availability(self.appliances[jid], True)            
        except KeyError:
            await self._add_appliance(jid)
            self.appliances[jid].set_available()

    async def _on_presence_unavailable(self, evt):
        """When appliance is no longer available, mark it as such."""
        jid = slixmpp.JID(evt['from']).bare

        if jid == self._client.boundjid.bare:
            return
        try:
            self._set_appliance_availability(self.appliances[jid], False)
        except KeyError:
            pass

    async def _on_message(self, event):
        """Handle incoming messages."""

        msg = str(event)
        msg_from = slixmpp.JID(event['from']).bare
        try:
            message_data = self._extract_message_json(msg)
        except ValueError:
            _LOGGER.info(f"From: {msg_from}: Not a GE message")
            return
        try:
            appliance = self.appliances[msg_from]
            state_changes = appliance.update_erd_values(message_data)
            if state_changes:
                await self.async_event(EVENT_APPLIANCE_STATE_CHANGE, [appliance, state_changes])
            await self.async_event(EVENT_APPLIANCE_UPDATE_RECEIVED, [appliance, message_data])
        except KeyError:
            _LOGGER.warning('Received json message from unregistered appliance')

    async def _on_start(self, event):
        _LOGGER.debug('Starting session: ' + str(event))
        self._client.send_presence('available')
        await asyncio.sleep(5)
        await self._set_connected()
        await self.async_event(EVENT_GOT_APPLIANCE_LIST, None)

    def _complete_jid(self, jid) -> str:
        """Make a complete jid from a username"""
        if "@" in jid:
            return jid
        return f"{jid}@{self._client.boundjid.host}"

    def _get_appliance_jid(self, appliance: GeAppliance) -> str:
        return self._complete_jid(f"{appliance.mac_addr}_{self.user_id}")

    def _send_raw_message(self, mto: slixmpp.JID, mbody: str, mtype: str = 'chat', msg_id: Optional[str] = None):
        """TODO: Use actual xml for this instead of hacking it.  Then again, this is what GE does in the app."""
        try:
            if not self._client:
                raise RuntimeError('Client connection is not available')

            if msg_id is None:
                msg_id = self._client.new_id()
            raw_message = (
                f'<message type="{mtype}" from="{self._client.boundjid.bare}" to="{mto}" id="{msg_id}">'
                f'<body>{mbody}</body>'
                f'</message>'
            )
            
            self._client.send_raw(raw_message)
        except Exception as err:
            raise GeRequestError from err

    def _send_request(
            self, appliance: GeAppliance, method: str, uri: str, key: Optional[str] = None,
            value: Optional[str] = None, message_id: Optional[str] = None):
        """
        Send a pseudo-http request to the appliance
        :param appliance: GeAppliance, the appliance to send the request to
        :param method: str, Usually "GET" or "POST"
        :param uri: str, the "endpoint" for the request, usually of the form "/UUID/erd/{erd_code}"
        :param key: The json key to set in a POST request.  Usually a four-character hex string with leading "0x".
        :param value: The value to set, usually encoded as a hex string without a leading "0x"
        :param message_id:
        """
        if method.lower() != 'post' and (key is not None or value is not None):
            raise RuntimeError('Values can only be set in a POST request')

        if message_id is None:
            message_id = self._client.new_id()
        message_body = self._format_request(message_id, uri, method, key, value)
        jid = slixmpp.JID(self._get_appliance_jid(appliance))
        self._send_raw_message(jid, message_body)

    async def async_request_update(self, appliance: GeAppliance):
        """
        Request a full update from the appliance.
        TODO: This doesn't seem to do a full request.  Need to investigate if this is possible.
        """
        self._send_request(appliance, 'GET', '/UUID/cache')

    async def async_set_erd_value(self, appliance: GeAppliance, erd_code: ErdCodeType, erd_value: Any):
        """
        Send a new erd value to the appliance
        :param appliance: GeAppliance, the appliance to update
        :param erd_code: The ERD code to update
        :param erd_value: The new value to set
        """
        if isinstance(erd_code, ErdCode):
            raw_erd_code = erd_code.value
        else:
            raw_erd_code = erd_code

        uri = f'/UUID/erd/{raw_erd_code}'
        self._send_request(appliance, 'POST', uri, raw_erd_code, erd_value)

    @staticmethod
    def _extract_message_json(message: str) -> Dict:
        """The appliances send messages that don't play nice with slixmpp, so let's do this."""
        etr = etree.XML(message)
        json_elem = _first_or_none(etr.xpath('//json'))
        if json_elem is None:
            raise ValueError('Not a GE appliance message')

        data = _first_or_none(json_elem.xpath('text()'))
        data = json.loads(data)
        return data

    @staticmethod
    def _format_request(
            msg_id: Union[int, str], uri: str, method: str, key: Optional[str] = None, value: Optional[str] = None) -> str:
        """Format a XMPP pseudo-HTTP request."""
        if method.lower() == 'post':
            post_body = f"<json>{json.dumps({key:value})}</json>"
        else:
            post_body = ""
        message = (
            f"<request><id>{msg_id}</id>"
            f"<method>{method}</method>"
            f"<uri>{uri}</uri>"
            f"{post_body}"
            "</request>"  # [sic.]
        )
        return message



from .erd_cycle_state import ErdCycleState, ErdCycleStateRaw

CYCLE_STATE_RAW_MAP = {
    ErdCycleStateRaw.PREWASH: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.PREWASH1: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.AUTO_HOT_START1: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.AUTO_HOT_START2: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.AUTO_HOT_START3: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.END_PREWASH1: ErdCycleState.PRE_WASH,
    ErdCycleStateRaw.SENSING: ErdCycleState.SENSING,
    ErdCycleStateRaw.MAIN_WASH: ErdCycleState.MAIN_WASH,
    ErdCycleStateRaw.DIVERTER_CAL: ErdCycleState.MAIN_WASH,
    ErdCycleStateRaw.DRYING: ErdCycleState.DRYING,
    ErdCycleStateRaw.SANITIZING: ErdCycleState.SANITIZING,
    ErdCycleStateRaw.RINSING: ErdCycleState.RINSING,
    ErdCycleStateRaw.TURNIDITY_CAL: ErdCycleState.RINSING,
    ErdCycleStateRaw.FINAL_RINSE: ErdCycleState.RINSING,
    ErdCycleStateRaw.FINAL_RINSE_FILL: ErdCycleState.RINSING,
    ErdCycleStateRaw.PAUSE: ErdCycleState.PAUSE,
    ErdCycleStateRaw.STATE_17: ErdCycleState.NA,
    ErdCycleStateRaw.STATE_18: ErdCycleState.NA,
    ErdCycleStateRaw.CYCLE_INACTIVE: ErdCycleState.NA,
    ErdCycleStateRaw.MAX: ErdCycleState.NA,
    ErdCycleStateRaw.INVALID: ErdCycleState.NA
}



from aiohttp import BasicAuth, ClientSession
from lxml import etree
from urllib.parse import urlparse, parse_qs

from ..exception import *
from .const import (
    LOGIN_URL, 
    OAUTH2_CLIENT_ID, 
    OAUTH2_CLIENT_SECRET,
    OAUTH2_REDIRECT_URI
)

try:
    import re2 as re
except ImportError:
    import re

async def async_get_oauth2_token(session: ClientSession, account_username: str, account_password: str):
    """Hackily get an oauth2 token until I can be bothered to do this correctly"""

    params = {
        'client_id': OAUTH2_CLIENT_ID,
        'response_type': 'code',
        'access_type': 'offline',
        'redirect_uri': OAUTH2_REDIRECT_URI,
    }

    async with session.get(f'{LOGIN_URL}/oauth2/auth', params=params) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        resp_text = await resp.text()

    email_regex = (
        r'^\s*(\w+(?:(?:-\w+)|(?:\.\w+)|(?:\+\w+))*\@'
        r'[A-Za-z0-9]+(?:(?:\.|-)[A-Za-z0-9]+)*\.[A-Za-z0-9][A-Za-z0-9]+)\s*$'
    )
    clean_username = re.sub(email_regex, r'\1', account_username)

    etr = etree.HTML(resp_text)
    post_data = {
        i.attrib['name']: i.attrib['value']
        for i in etr.xpath("//form[@id = 'frmsignin']//input")
        if 'value' in i.keys()
    }
    post_data['username'] = clean_username
    post_data['password'] = account_password

    async with session.post(f'{LOGIN_URL}/oauth2/g_authenticate', data=post_data, allow_redirects=False) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        code = parse_qs(urlparse(resp.headers['Location']).query)['code'][0]

    post_data = {
        'code': code,
        'client_id': OAUTH2_CLIENT_ID,
        'client_secret': OAUTH2_CLIENT_SECRET,
        'redirect_uri': OAUTH2_REDIRECT_URI,
        'grant_type': 'authorization_code',
    }
    auth = BasicAuth(OAUTH2_CLIENT_ID, OAUTH2_CLIENT_SECRET)
    async with session.post(f'{LOGIN_URL}/oauth2/token', data=post_data, auth=auth) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        oauth_token = await resp.json()
    try:
        access_token = oauth_token['access_token']
        return oauth_token
    except KeyError:
        raise GeAuthFailedError(f'Failed to get a token: {oauth_token}')

async def async_refresh_oauth2_token(session: ClientSession, refresh_token: str):
    """ Refreshes an OAuth2 Token based on a refresh token """

    post_data = {
        'redirect_uri': OAUTH2_REDIRECT_URI,
        'client_id': OAUTH2_CLIENT_ID,
        'client_secret': OAUTH2_CLIENT_SECRET,
        'grant_type': 'refresh_token',
        'refresh_token': refresh_token
    }
    auth = BasicAuth(OAUTH2_CLIENT_ID, OAUTH2_CLIENT_SECRET)
    async with session.post(f'{LOGIN_URL}/oauth2/token', data=post_data, auth=auth) as resp:
        if 400 <= resp.status < 500:
            raise GeAuthFailedError(await resp.text())
        if resp.status >= 500:
            raise GeGeneralServerError(await resp.text())
        oauth_token = await resp.json()
    try:
        access_token = oauth_token['access_token']
        return oauth_token
    except KeyError:
        raise GeAuthFailedError(f'Failed to get a token: {oauth_token}')



import asyncio
from collections import defaultdict
import os
import logging
import socket
import slixmpp
import ssl
from typing import Any, Dict, List, Callable

try:
    import ujson as json
except ImportError:
    import json

_LOGGER = logging.getLogger(__name__)

# If this isn't done, it'll throw a not implemented exception
if os.name == 'nt':
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

XMPP_EVENT_MESSAGE = 'message'
XMPP_EVENT_SESSION_START = 'session_start'
XMPP_EVENT_PRESENCE_AVAILABLE = 'presence_available'
XMPP_EVENT_PRESENCE_UNAVAILABLE = 'presence_unavailable'

class GeClientXMPP(slixmpp.ClientXMPP):
    def __init__(self, jid, password):
        super().__init__(jid, password)

        self.add_event_handler(XMPP_EVENT_MESSAGE, self.on_message)
        self.add_event_handler(XMPP_EVENT_SESSION_START, self.on_start)
        self.add_event_handler(XMPP_EVENT_PRESENCE_AVAILABLE, self.on_presence_available)
        self.add_event_handler(XMPP_EVENT_PRESENCE_UNAVAILABLE, self.on_presence_unavailable)
        self.ssl_context.verify_mode = ssl.CERT_NONE

        self._event_handlers = defaultdict(list)

    @property
    def event_handlers(self) -> Dict[str, List[Callable]]:
        return self._event_handlers

    async def async_event(self, event: str, *args, **kwargs):
        """Trigger event callbacks sequentially"""
        for cb in self.event_handlers[event]:
            asyncio.ensure_future(cb(*args, **kwargs), loop=self.loop)

    def add_external_event_handler(self, event: str, callback: Callable, disposable: bool = False):
        if disposable:
            raise NotImplementedError('Support for disposable callbacks not yet implemented')
        self.event_handlers[event].append(callback)

    async def on_presence_available(self, evt: slixmpp.ElementBase):
        await self.async_event(XMPP_EVENT_PRESENCE_AVAILABLE, evt)
    
    async def on_presence_unavailable(self, evt: slixmpp.ElementBase):
        await self.async_event(XMPP_EVENT_PRESENCE_UNAVAILABLE, evt)

    async def on_start(self, evt):
        await self.async_event(XMPP_EVENT_SESSION_START, evt)

    async def on_message(self, evt):
        await self.async_event(XMPP_EVENT_MESSAGE, evt)

    async def _connect_routine(self):
        """
        Override the _connect_routine method from xmlstream.py.
        This is to address the bug corrected in open PR https://github.com/poezio/slixmpp/pull/19.
        """
        self.event_when_connected = "connected"

        if self.connect_loop_wait > 0:
            self.event('reconnect_delay', self.connect_loop_wait)
            await asyncio.sleep(self.connect_loop_wait, loop=self.loop)

        record = await self.pick_dns_answer(self.default_domain)
        if record is not None:
            host, address, dns_port = record
            port = self.address[1] if self.address[1] else dns_port
            self.address = (address, port)
            self._service_name = host
        else:
            # No DNS records left, stop iterating
            # and try (host, port) as a last resort
            self.dns_answers = None

        if self.use_ssl:
            ssl_context = self.get_ssl_context()
        else:
            ssl_context = None

        if self._current_connection_attempt is None:
            return
        try:
            await self.loop.create_connection(
                lambda: self, self.address[0], self.address[1], ssl=ssl_context,
                server_hostname=self.default_domain if self.use_ssl else None
            )
            self.connect_loop_wait = 0
        except socket.gaierror:
            self.event('connection_failed', 'No DNS record available for %s' % self.default_domain)
        except OSError as e:
            _LOGGER.debug('Connection failed: %s', e)
            self.event("connection_failed", e)
            if self._current_connection_attempt is None:
                return
            self.connect_loop_wait = self.connect_loop_wait * 2 + 1
            self._current_connection_attempt = asyncio.ensure_future(
                self._connect_routine(), loop=self.loop,
            )



"""Base client for GE ERD APIs"""

import abc
from gekitchensdk.clients.async_login_flows import async_get_oauth2_token, async_refresh_oauth2_token
from aiohttp import ClientSession
import asyncio
from collections import defaultdict
from datetime import datetime, timedelta
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple

from ..erd import ErdCode, ErdCodeType
from ..exception import *
from ..ge_appliance import GeAppliance
from .const import (
    EVENT_APPLIANCE_INITIAL_UPDATE,
    EVENT_APPLIANCE_AVAILABLE,
    EVENT_APPLIANCE_UNAVAILABLE, 
    EVENT_CONNECTED, 
    EVENT_DISCONNECTED, 
    EVENT_STATE_CHANGED,
    MAX_RETRIES, 
    RETRY_INTERVAL,
)
from .states import GeClientState

_LOGGER = logging.getLogger(__name__)

class GeBaseClient(metaclass=abc.ABCMeta):
    """Abstract base class for GE ERD APIs"""

    client_priority = 0  # Priority of this client class.  Higher is better.

    def __init__(self, username: str, password: str, event_loop: Optional[asyncio.AbstractEventLoop] = None):
        self.account_username = username
        self.account_password = password
        self._credentials = None  # type: Optional[Dict]
        self._session = None # type: Optional[ClientSession]

        self._access_token = None
        self._refresh_token = None
        self._token_expiration_time = datetime.now()

        self._state = GeClientState.INITIALIZING
        self._disconnect_requested = asyncio.Event()
        self._retries_since_last_connect = -1
        self._has_successful_connect = False
        self._loop = event_loop
        self._appliances = {}  # type: Dict[str, GeAppliance]
        self._initialize_event_handlers()

    @property
    def credentials(self) -> Optional[Dict]:
        return self._credentials

    @credentials.setter
    def credentials(self, credentials: Dict):
        self._credentials = credentials

    @property
    def appliances(self) -> Dict[str, GeAppliance]:
        return self._appliances

    @property
    def user_id(self) -> Optional[str]:
        try:
            return self.credentials['userId']
        except (TypeError, KeyError):
            raise GeNotAuthenticatedError

    @property
    def state(self) -> GeClientState:
        return self._state

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
        return self._loop

    @property
    def connected(self) -> bool:
        """ Indicates whether the client is in a connected state """
        return self._state not in [GeClientState.DISCONNECTING, GeClientState.DISCONNECTED]
    
    @property
    def available(self) -> bool:
        """ Indicates whether the client is available for sending/receiving commands """
        return self._state == GeClientState.CONNECTED

    @property
    def event_handlers(self) -> Dict[str, List[Callable]]:
        return self._event_handlers

    async def async_event(self, event: str, *args, **kwargs):
        """Trigger event callbacks sequentially"""
        for cb in self.event_handlers[event]:
            asyncio.ensure_future(cb(*args, **kwargs), loop=self.loop)

    def add_event_handler(self, event: str, callback: Callable, disposable: bool = False):
        if disposable:
            raise NotImplementedError('Support for disposable callbacks not yet implemented')
        self.event_handlers[event].append(callback)

    def remove_event_handler(self, event: str, callback: Callable):
        try:
            self.event_handlers[event].remove(callable)
        except:
            _LOGGER.warn(f"could not remove event handler {event}-{callable}")

    def clear_event_handlers(self):
        self._initialize_event_handlers()

    async def async_get_credentials_and_run(self, session: ClientSession):
        """Do a full login flow and run the client."""
        await self.async_get_credentials(session)
        await self.async_run_client()

    async def async_run_client(self):
        #reset the disconnect event
        self._disconnect_requested.clear()

        _LOGGER.info('Starting GE Appliances client')
        while not self._disconnect_requested.is_set():
            if self._retries_since_last_connect > MAX_RETRIES:
                _LOGGER.debug(f'Tried auto-reconnecting {MAX_RETRIES} times, giving up.')
                break
            try:
                await self._async_run_client()
            except GeNeedsReauthenticationError:
                _LOGGER.info('Reauthentication needed')
            except GeRequestError as err:
                _LOGGER.info(f'Error executing request {err}')
            except Exception as err:
                if not self._has_successful_connect:
                    _LOGGER.warn(f'Unhandled exception on first connect attempt: {err}, disconnecting')
                    break
                _LOGGER.info(f'Unhandled exception while running client: {err}, ignoring and restarting')  
            finally:
                if not self._disconnect_requested.is_set():
                    await self._set_state(GeClientState.DROPPED)
                    await self._set_state(GeClientState.WAITING)
                    _LOGGER.debug('Waiting before reconnecting')
                    await asyncio.sleep(RETRY_INTERVAL)
                    _LOGGER.debug('Refreshing authentication before reconnecting')
                    try:
                        await self.async_do_refresh_login_flow()
                    except Exception as err:
                        #if there was an error refreshing the authentication, break the loop and kill the client
                        _LOGGER.warn(f'Error refreshing authentication: {err}')
                        break
                self._retries_since_last_connect += 1

        #initiate the disconnection            
        await self.disconnect()

    @abc.abstractmethod
    async def _async_run_client(self):
        """ Internal method to run the client """

    @abc.abstractmethod
    async def async_set_erd_value(self, appliance: GeAppliance, erd_code: ErdCodeType, erd_value: Any):
        """
        Send a new erd value to the appliance
        :param appliance: The appliance being updated
        :param erd_code: The ERD code to update
        :param erd_value: The new value to set
        """
        pass

    @abc.abstractmethod
    async def async_request_update(self, appliance: GeAppliance):
        """Request the appliance send a full state update"""
        pass

    async def async_get_credentials(self, session: ClientSession):
        """Get updated credentials"""
        self._session = session
        await self.async_do_full_login_flow()
        
    async def async_do_full_login_flow(self) -> Dict[str, str]:
        """Do the full login flow for this client"""
        self.credentials = await self._async_do_full_login_flow()
        return self.credentials

    @abc.abstractmethod
    async def _async_do_full_login_flow(self) -> Dict[str, str]:
        """Internal full login flow"""
        pass

    async def async_do_refresh_login_flow(self) -> Dict[str, str]:
        """Do the refresh login flow for this client"""
        self.credentials = await self._async_do_refresh_login_flow()
        return self.credentials

    @abc.abstractmethod
    async def _async_do_refresh_login_flow(self) -> Dict[str, str]:
        """Internal refresh login flow"""
        pass

    async def _async_get_oauth2_token(self):
        """Get the OAuth2 token based on the username and password"""

        await self._set_state(GeClientState.AUTHORIZING_OAUTH)

        oauth_token = await async_get_oauth2_token(self._session, self.account_username, self.account_password)

        try:
            self._access_token = oauth_token['access_token']
            self._token_expiration_time = datetime.now() + timedelta(seconds=(oauth_token['expires_in'] - 120))
            self._refresh_token = oauth_token['refresh_token']
        except KeyError:
            raise GeAuthFailedError(f'Failed to get a token: {oauth_token}')

    async def _async_refresh_oauth2_token(self):
        """ Refreshes an OAuth2 Token based on a refresh token """

        await self._set_state(GeClientState.AUTHORIZING_OAUTH)

        oauth_token = await async_refresh_oauth2_token(self._session, self._refresh_token)

        try:
            self._access_token = oauth_token['access_token']
            self._token_expiration_time = datetime.now() + timedelta(seconds=(oauth_token['expires_in'] - 120))
            self._refresh_token = oauth_token.get('refresh_token', self._refresh_token)
        except KeyError:
            raise GeAuthFailedError(f'Failed to get a token: {oauth_token}')

    async def _maybe_trigger_appliance_init_event(self, data: Tuple[GeAppliance, Dict[ErdCodeType, Any]]):
        """
        Trigger the appliance_got_type event if appropriate

        :param data: GeAppliance updated and the updates
        """
        appliance, state_changes = data
        if ErdCode.APPLIANCE_TYPE in state_changes and not appliance.initialized:
            _LOGGER.debug(f'Got initial appliance type for {appliance:s}')
            appliance.initialized = True
            await self.async_event(EVENT_APPLIANCE_INITIAL_UPDATE, appliance)

    async def _set_appliance_availability(self, appliance: GeAppliance, available: bool):
        if available and not appliance.available:
            appliance.set_available()
            await self.async_event(EVENT_APPLIANCE_AVAILABLE, appliance)
        elif not available and appliance.available:
            appliance.set_unavailable()
            await self.async_event(EVENT_APPLIANCE_UNAVAILABLE, appliance)

    async def _set_state(self, new_state: GeClientState) -> bool:
        """ Indicate that the state changed and raise an event """
        if self._state != new_state:
            old_state = self._state
            self._state = new_state
            await self.async_event(EVENT_STATE_CHANGED, old_state, new_state)
            return True
        return False            

    def _initialize_event_handlers(self):
        self._event_handlers = defaultdict(list)  # type: Dict[str, List[Callable]]
        self.add_event_handler(EVENT_STATE_CHANGED, self._on_state_change)
        pass

    async def _on_state_change(self, old_state: GeClientState, new_state: GeClientState):
        _LOGGER.debug(f'Client changed state: {old_state} to {new_state}')

        if new_state == GeClientState.CONNECTED:
            await self.async_event(EVENT_CONNECTED, None)
        if new_state == GeClientState.DISCONNECTED:
            await self.async_event(EVENT_DISCONNECTED, None)

    async def disconnect(self):
        """Disconnect and cleanup."""
        if not self._disconnect_requested.is_set():
            _LOGGER.info("Disconnecting")
            await self._set_state(GeClientState.DISCONNECTING)         
            self._disconnect_requested.set()
            await self._disconnect()
            await self._set_state(GeClientState.DISCONNECTED) 

    async def _set_connected(self):
        self._retries_since_last_connect = -1
        self._has_successful_connect = True
        await self._set_state(GeClientState.CONNECTED)

    @abc.abstractmethod
    async def _disconnect(self) -> None:
        pass




import asyncio
from typing import AsyncIterator


async def CancellableAsyncIterator(async_iterator: AsyncIterator, cancellation_event: asyncio.Event) -> AsyncIterator:
    cancellation_task = asyncio.create_task(cancellation_event.wait())
    result_iter = async_iterator.__aiter__()
    while not cancellation_event.is_set():
        done, pending = await asyncio.wait(
            [cancellation_task, result_iter.__anext__()],
            return_when=asyncio.FIRST_COMPLETED
        )
        for done_task in done:
            if done_task == cancellation_task:
                for pending_task in pending:
                    await pending_task
                break
            else:
                yield done_task.result()



import enum

@enum.unique
class GeClientState(enum.Enum):
    INITIALIZING = enum.auto()
    AUTHORIZING_OAUTH = enum.auto()
    AUTHORIZING_CLIENT = enum.auto()
    CONNECTING = enum.auto()
    CONNECTED = enum.auto()
    DROPPED = enum.auto()
    WAITING = enum.auto()
    DISCONNECTING = enum.auto()
    DISCONNECTED = enum.auto()



from .ge_exception import GeException

class GeRequestError(GeException):
    """ Exception raised when there is an error processing a message """
    def __init__(self, message: str, code, reason, *args: object) -> None:
        super().__init__(*args)
        self.code = code
        self.reason = reason
        self.message = message
    
    def __str__(self) -> str:
        return f"There was an error while processing a message: Code={self.code}, Reason={self.reason}, Message={self.message}"




"""GE Client implementations"""

import logging
from .const import (
    EVENT_ADD_APPLIANCE,
    EVENT_APPLIANCE_INITIAL_UPDATE,
    EVENT_APPLIANCE_STATE_CHANGE,
    EVENT_APPLIANCE_AVAILABLE,
    EVENT_APPLIANCE_UNAVAILABLE,
    EVENT_APPLIANCE_UPDATE_RECEIVED,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    EVENT_GOT_APPLIANCE_LIST,
    EVENT_STATE_CHANGED 
)
from .base_client import GeBaseClient
from .websocket_client import GeWebsocketClient
from .async_login_flows import async_get_oauth2_token, async_refresh_oauth2_token

_LOGGER = logging.getLogger(__name__)

try:
    from .xmpp_client import GeXmppClient
except ImportError:
    _LOGGER.info("XMPP client not avaible.  You may need to install slximpp.")



import enum

@enum.unique
class ErdCycleStateRaw(enum.Enum):
    NO_CHANGE = 0
    PREWASH = 1
    SENSING = 2
    MAIN_WASH = 3
    DRYING = 4
    SANITIZING = 5
    TURNIDITY_CAL = 6
    DIVERTER_CAL = 7
    PAUSE = 8
    RINSING = 9
    PREWASH1 = 10
    FINAL_RINSE = 11
    END_PREWASH1 = 12
    AUTO_HOT_START1 = 13
    AUTO_HOT_START2 = 14
    AUTO_HOT_START3 = 15
    FINAL_RINSE_FILL = 16
    STATE_17 = 17
    STATE_18 = 18
    CYCLE_INACTIVE = 26
    MAX = 27
    INVALID = 255

@enum.unique
class ErdCycleState(enum.Enum):
    NA = -1
    PRE_WASH = 1
    SENSING = 2
    MAIN_WASH = 3
    DRYING = 4
    SANITIZING = 5
    RINSING = 6
    PAUSE = 7
